package ru.dbotthepony.kstarbound.api

import java.io.*
import java.nio.ByteBuffer

interface IVFS {
	fun pathExists(path: String): Boolean

	fun pathExistsOrElse(path: String, orElse: String): String {
		if (pathExists(path))
			return path

		return orElse
	}

	fun firstExisting(vararg pathList: String): String {
		for (path in pathList)
			if (pathExists(path))
				return path

		throw FileNotFoundException("Unable to find any of files specified")
	}

	fun firstExistingOrNull(vararg pathList: String): String? {
		for (path in pathList)
			if (pathExists(path))
				return path

		return null
	}

	fun read(path: String): ByteBuffer {
		return readOrNull(path) ?: throw FileNotFoundException("No such file $path")
	}

	fun readOrNull(path: String): ByteBuffer?

	fun readString(path: String): String {
		return read(path).array().toString(Charsets.UTF_8)
	}

	fun getReader(path: String): Reader {
		return InputStreamReader(ByteArrayInputStream(read(path).array()))
	}

	fun listFiles(path: String): Collection<String>
	fun listDirectories(path: String): Collection<String>

	fun listFilesAndDirectories(path: String): Collection<String> {
		val a = listFiles(path)
		val b = listDirectories(path)

		return ArrayList<String>(a.size + b.size).also { it.addAll(a); it.addAll(b) }
	}

	fun listAllFilesWithExtension(extension: String): Collection<String> {
		val listing = ArrayList<String>()
		val ext = ".$extension"

		for (listedFile in listAllFiles("")) {
			if (listedFile.endsWith(ext)) {
				listing.add(listedFile)
			}
		}

		return listing
	}

	fun listAllFiles(path: String): Collection<String> {
		val lists = mutableListOf<Collection<String>>()

		lists.add(listFiles(path))

		for (dir in listDirectories(path)) {
			lists.add(listAllFiles(dir))
		}

		// flatten медленный
		// return lists.flatten()

		var size = 0

		for (list in lists) {
			size += list.size
		}

		return ArrayList<String>(size).also { lists.forEach(it::addAll) }
	}

	fun readDirect(path: String): ByteBuffer {
		val read = read(path)

		val buf = ByteBuffer.allocateDirect(read.capacity())

		read.position(0)

		for (i in 0 until read.capacity()) {
			buf.put(read[i])
		}

		buf.position(0)

		return buf
	}
}

fun getPathFolder(path: String): String {
	return path.substringBeforeLast('/')
}

fun getPathFilename(path: String): String {
	return path.substringAfterLast('/')
}

class PhysicalFS(root: File) : IVFS {
	val root: File = root.absoluteFile

	override fun pathExists(path: String): Boolean {
		if (path.contains("..")) {
			return false
		}

		return File(root.absolutePath, path).exists()
	}

	override fun readOrNull(path: String): ByteBuffer? {
		if (path.contains("..")) {
			return null
		}

		val fpath = File(root.path, path)

		if (!fpath.exists()) {
			return null
		}

		val reader = fpath.inputStream()
		val buf = ByteBuffer.allocate(reader.channel.size().toInt())
		reader.read(buf.array())
		return buf
	}

	override fun listFiles(path: String): List<String> {
		if (path.contains("..")) {
			return listOf()
		}

		val fpath = File(root.absolutePath, path)
		return fpath.listFiles()?.filter { it.isFile }?.map {
			it.path.replace('\\', '/').substring(root.path.length)
		} ?: return listOf()
	}

	override fun listDirectories(path: String): Collection<String> {
		if (path.contains("..")) {
			return listOf()
		}

		val fpath = File(root.absolutePath, path)
		return fpath.listFiles()?.filter { it.isDirectory }?.map {
			it.path.replace('\\', '/').substring(root.path.length)
		} ?: return listOf()
	}
}
